/** simext.c Copyright 1992-2019 Ventana Systems, Inc.
**
**  ABSTRACT:
**      This file contains the supporting code for directly called 
        functions used - or potentially used - in mdl.c
**  AUTHORS:
**
**      Ventana Systems, Inc. http://www.vensim.com
**/
/*******************************************************************
  COMPILER NOTES

  This is a standard C program and should work with any compiler though 
  it is only tested with Microsoft C.  

  To use this with the double precision version of Vensim it is 
  necessary to define DBLPREC and recompile.  

  
********************************************************************/

#pragma pack(push,4)
#define _x64
#define BUILD_x64

#include "vensim.h"


/*************************************************************************

  Part 2

  Function prototypes and macro definitions for use in compiled simulations.
  Note that you can call the exported functions listed here from external
  functions as well if that is at all helpful.

  ********************************************************************/


#define TINY 1e-6 
/* functions defined as macros to work more easily */ 
#define FIRST_IF_TRUE(a,b,c) ((a)?(b):(c)) 
#define IF_THEN_ELSE(a,b,c) ((a)?(b):(c)) 
//#define ABS(a) ((a)>0?(a):-(a)) 
#define ABS(a) fabs(a) /* single evaluation of the argument */
#ifndef SIMEXT
//#define MAX(a,b) ((a) > (b)?(a):(b)) 
BigREAL MAX(BigREAL a,BigREAL b)
 {
  if(a > b)
     return a ;
  return b ;
  }
//#define MIN(a,b) ((a) > (b)?(b):(a)) 
BigREAL MIN(BigREAL a,BigREAL b)
 {
  if(a < b)
     return a ;
  return b ;
  }
#endif
 

#define setpos(a) VGV->lastpos = a ; // todo: this could be nulled out for speed, except that TABLE requires VGV->lastpos
#define setposF(a) VGV->lastpos = a ; // this is always needed (for :EXCEPT:)



#define LOG(a,b) (log(a)/log(b)) 
#define LN(a)  (log(a)) 
/* XIDZ and ZIDZ need to behave the same as for interpreted which means we need
   an extra argument and a max nesting depth (set at 10) */
#define XIDZ(n,a,b,c) ((temp[n]=(b)) <= TINY && temp[n] > -TINY?(c):(a)/temp[n])
#define ZIDZ(n,a,b) ((temp[n]=(b)) <= TINY && temp[n] > -TINY?0:(a)/temp[n])
#define EXP(a) exp(a) 
#define COS(a) cos(a) 
#define COSH(a) cosh(a) 
#define SIN(a) sin(a) 
#define SINH(a) sinh(a) 
#define ARCCOS(a) acos(a) 
#define ARCSIN(a) asin(a) 
#define ARCTAN(a) atan(a) 
#define TAN(a) tan(a) 
#define TANH(a) tanh(a) 
#define SQRT(a) sqrt(a) 
#define INTEGER(a) ((COMPREAL)((long)(a)))

/* functions used in models */ 
CFUNCTION BigREAL TABLE(TAB_TYPE *tab,BigREAL in) ; 
CFUNCTION BigREAL LOOKUP_BACKWARD(TAB_TYPE *tab,BigREAL in) ; 
CFUNCTION BigREAL LOOKUP_FORWARD(TAB_TYPE *tab,BigREAL in) ; 
CFUNCTION BigREAL LOOKUP_INVERT(TAB_TYPE *tab,BigREAL in) ; 
CFUNCTION BigREAL LOOKUP_EXTRAPOLATE(TAB_TYPE *tab,BigREAL in) ; 
CFUNCTION BigREAL LOOKUP_AREA(TAB_TYPE *tab,BigREAL start,BigREAL end) ; 
CFUNCTION BigREAL LOOKUP_SLOPE(TAB_TYPE *tab,BigREAL x,BigREAL mode) ; 
CFUNCTION BigREAL MESSAGE(char *msg,BigREAL severity) ;
CFUNCTION BigREAL RANDOM_0_1 (void) ; 
CFUNCTION BigREAL RANDOM_BETA(BigREAL xmin,BigREAL xmax,BigREAL alpha,BigREAL beta,BigREAL shift,BigREAL stretch,BigREAL streamid) ;
CFUNCTION BigREAL RANDOM_BINOMIAL(BigREAL xmin,BigREAL xmax,BigREAL p,BigREAL n,BigREAL shift,BigREAL stretch,BigREAL streamid) ;
CFUNCTION BigREAL RANDOM_EXPONENTIAL(BigREAL xmin,BigREAL xmax,BigREAL shift,BigREAL stretch,BigREAL streamid) ;
CFUNCTION BigREAL RANDOM_GAMMA(BigREAL xmin,BigREAL xmax,BigREAL a,BigREAL shift,BigREAL stretch,BigREAL streamid) ;
CFUNCTION BigREAL RANDOM_LOOKUP(TAB_TYPE *look,BigREAL xmin,BigREAL xmax,BigREAL shift,BigREAL stretch,BigREAL streamid) ;
CFUNCTION BigREAL RANDOM_NEGATIVE_BINOMIAL(BigREAL xmin,BigREAL xmax,BigREAL p,BigREAL n,BigREAL shift,BigREAL stretch,BigREAL streamid) ;
CFUNCTION BigREAL RANDOM_NORMAL(BigREAL xmin,BigREAL xmax,BigREAL shift,BigREAL stretch,BigREAL streamid) ;
CFUNCTION BigREAL RANDOM_POISSON(BigREAL xmin,BigREAL xmax,BigREAL a,BigREAL shift,BigREAL stretch,BigREAL streamid) ;
CFUNCTION BigREAL RANDOM_TRIANGULAR(BigREAL xmin,BigREAL xmax,BigREAL peak,BigREAL shift,BigREAL stretch,BigREAL streamid) ;
CFUNCTION BigREAL RANDOM_UNIFORM(BigREAL xmin,BigREAL xmax,BigREAL streamid) ;
CFUNCTION BigREAL RANDOM_WEIBULL(BigREAL xmin,BigREAL xmax,BigREAL shape,BigREAL shift,BigREAL stretch,BigREAL streamid) ;
CFUNCTION BigREAL QUANTUM(BigREAL a,BigREAL b) ; 
CFUNCTION BigREAL MODULO(BigREAL a,BigREAL b) ; 
CFUNCTION BigREAL STEP(GLOB_VARS *VGV,BigREAL a,BigREAL b) ; 
CFUNCTION BigREAL GAMMA_LN(BigREAL a) ; 
CFUNCTION BigREAL RAMP(GLOB_VARS *VGV,BigREAL a,BigREAL b,BigREAL c) ; 
CFUNCTION BigREAL PULSE(GLOB_VARS *VGV,BigREAL ftime,BigREAL width) ; 
CFUNCTION BigREAL PULSE_TRAIN(GLOB_VARS *VGV,BigREAL  ftime,BigREAL  width,BigREAL repeattime,BigREAL stop) ;
CFUNCTION BigREAL SHIFT_IF_TRUE(GLOB_VARS *VGV,VECTOR_ARG *array,BigREAL cond,BigREAL arrsize,BigREAL cumulate,
                 BigREAL startval) ;
CFUNCTION BigREAL MARKETP(VECTOR_ARG *request,VECTOR_ARG *priority,int size,BigREAL width,BigREAL supply) ; 
CFUNCTION BigREAL ALLOCATE_BY_PRIORITY(GLOB_VARS *VGV,VECTOR_ARG *allocation,VECTOR_ARG *request,VECTOR_ARG *priority,int size,BigREAL width,BigREAL supply) ; 

CFUNCTION BigREAL ALLOCATE_AVAILABLE(VECTOR_ARG *allocation,VECTOR_ARG *request,VECTOR_ARG *request_p,BigREAL available) ;
CFUNCTION BigREAL MARKET_MATCH(VECTOR_ARG *sales,VECTOR_ARG *demandq,VECTOR_ARG *dshape,VECTOR_ARG *seller_attract,VECTOR_ARG *supplyq,VECTOR_ARG *sshape,VECTOR_ARG *buyer_priority,VECTOR_ARG *dqsconversion) ;
CFUNCTION BigREAL FIND_MARKET_PRICE(VECTOR_ARG *d,VECTOR_ARG *d_p,VECTOR_ARG *s,VECTOR_ARG *sp) ;
/* DEMAND AT PRICE and SUPPLY AT PRICE are really the same function */
#define DEMAND_AT_PRICE(ds,ds_p,p) MARKET_DS(ds,ds_p,p,1)
#define SUPPLY_AT_PRICE(ds,ds_p,p) MARKET_DS(ds,ds_p,p,0)
CFUNCTION BigREAL MARKET_DS(VECTOR_ARG *ds,VECTOR_ARG *ds_p,BigREAL price,int isdemand) ;

CFUNCTION BigREAL VECTOR_REORDER(VECTOR_ARG *outvector,VECTOR_ARG *invector,VECTOR_ARG *sortvector) ;
CFUNCTION BigREAL VECTOR_SORT_ORDER(VECTOR_ARG *outvector,VECTOR_ARG *invector,BigREAL d) ;
CFUNCTION BigREAL VECTOR_RANK(VECTOR_ARG *outvector,VECTOR_ARG *invector,BigREAL d) ;
CFUNCTION BigREAL ALLOC_P(BigREAL request,BigREAL priority,BigREAL width,BigREAL mp) ; 
CFUNCTION BigREAL TIME_BASE(GLOB_VARS *VGV,BigREAL intecept,BigREAL slope) ; 
CFUNCTION BigREAL INIT_PURE_DELAY(int type,VAL_OFF deloff,VAL_OFF valoff,double deltime,
                   double initval,double missval_or_order,TAB_TYPE *tab,TAB_TYPE *tab2) ;
CFUNCTION BigREAL STORE_PURE_DELAY(VAL_OFF deloff,double val,double outval,
        double delaytime,double changerate,TAB_TYPE *tab) ;
CFUNCTION BigREAL INIT_SINTEG(int type,
                   VAL_OFF deloff,VAL_OFF valoff,double initval,
                   double minval,double maxval,double quantval,
                   double specval,double tolerance) ;
CFUNCTION BigREAL SAMPLE_IF_TRUE(double oldval,double cond,double newval) ;
CFUNCTION BigREAL FIND_ZERO(VECTOR_ARG *vx,VECTOR_ARG *vsx,VECTOR_ARG *vz,VECTOR_ARG *vsz,VECTOR_ARG *inix,BigREAL fn,
      BigREAL fns,BigREAL tolerance,BigREAL fmax_iter,BigREAL method) ;
CFUNCTION BigREAL INVERT_MATRIX(VECTOR_ARG *ainv,VECTOR_ARG *a,BigREAL n) ;
CFUNCTION BigREAL POWER(BigREAL base,BigREAL exponent) ;
CFUNCTION BigREAL VECTOR_ELM_MAP(GLOB_VARS *VGV,VECTOR_ARG *in,BigREAL shift) ;
CFUNCTION BigREAL VECTOR_LOOKUP(VECTOR_ARG *in,BigREAL x,BigREAL xmin,BigREAL xmax,BigREAL mode) ;                 
CFUNCTION BigREAL GET_DATA_AT_TIME(VECTOR_ARG *v,BigREAL t) ;
CFUNCTION BigREAL GET_DATA_BETWEEN_TIMES(VECTOR_ARG *v,BigREAL t,BigREAL mode) ;
CFUNCTION BigREAL GET_DATA_ATTRIBUTE(int attrib,VECTOR_ARG *v,BigREAL t,BigREAL mode) ;
CFUNCTION BigREAL GET_DATA_FIRST_TIME(VECTOR_ARG *v) ;
CFUNCTION BigREAL GET_DATA_LAST_TIME(VECTOR_ARG *v) ;
CFUNCTION BigREAL GET_DATA_TOTAL_POINTS(VECTOR_ARG *v) ;
CFUNCTION BigREAL GET_TIME_VALUE(BigREAL relativeto,BigREAL offset,BigREAL measure) ;
CFUNCTION BigREAL QUEUE_AGE_AVERAGE(VECTOR_ARG *v,BigREAL x) ;
CFUNCTION BigREAL QUEUE_AGE_OLDEST(VECTOR_ARG *v) ;
CFUNCTION BigREAL QUEUE_AGE_IN_RANGE(VECTOR_ARG *v,BigREAL l,BigREAL h) ;
CFUNCTION BigREAL QUEUE_ATTRIB_AVERAGE(VECTOR_ARG *v,BigREAL x) ;
CFUNCTION BigREAL QUEUE_ATTRIB_MAX(VECTOR_ARG *v) ;
CFUNCTION BigREAL QUEUE_ATTRIB_MIN(VECTOR_ARG *v) ;
CFUNCTION BigREAL QUEUE_ATTRIB_IN_RANGE(VECTOR_ARG *v,BigREAL l,BigREAL h) ;
CFUNCTION BigREAL QUEUE_ATTRIB_QUANTITY(VECTOR_ARG *v,BigREAL x) ;

/* delay functions - defined as macros then passed back through
   an initial and active function */
#define DELAY_FIXED_i(type,varoff,delind,deltime,initval) INIT_PURE_DELAY(type,delind,varoff,deltime,initval,0.0,'\0','\0') 
#define DELAY_FIXED_a(delind,inval) STORE_PURE_DELAY(delind,inval,0.0,1.0,0,'\0')
#define DELAY_INFORMATION_i(type,varoff,delind,initval) INIT_PURE_DELAY(type,delind,varoff,1.0,initval,0.0,'\0','\0') 
#define DELAY_INFORMATION_a(delind,inval,deltime) STORE_PURE_DELAY(delind,inval,0.0,deltime,0,'\0')
#define DELAY_MATERIAL_i(type,varoff,delind,initval,missval) INIT_PURE_DELAY(type,delind,varoff,1.0,initval,missval,'\0','\0') 
#define DELAY_MATERIAL_a(delind,inval,deltime) STORE_PURE_DELAY(delind,inval,0.0,deltime,0,'\0')
/* these are not in 4.0f but will be going in future versions */
#define DELAY_N_i(type,varoff,delind,initval,order) INIT_PURE_DELAY(type,delind,varoff,1.0,initval,order,'\0','\0')
#define DELAY_N_a(delind,inval,deltime)  STORE_PURE_DELAY(delind,inval,0.0,deltime,0,'\0')
#define SMOOTH_N_i(type,varoff,delind,initval,order) INIT_PURE_DELAY(type,delind,varoff,1.0,initval,order,'\0','\0')
#define SMOOTH_N_a(delind,inval,deltime)  STORE_PURE_DELAY(delind,inval,0.0,deltime,0,'\0')
#define QUEUE_FIFO_i(type,varoff,delind,initprofile,initval,initage) INIT_PURE_DELAY(type,delind,varoff,initage,initval,0.0,initprofile,'\0')
#define QUEUE_FIFO_a(delind,inval,outval) STORE_PURE_DELAY(delind,inval,outval,1.0,0,'\0') ;
#define QUEUE_FIFO_ATTRIB_i(type,varoff,delind,initprofile,inattribprofile,initval,initattrib,initage) INIT_PURE_DELAY(type,delind,varoff,initage,initval,initattrib,initprofile,inattribprofile)
#define QUEUE_FIFO_ATTRIB_a(delind,inval,outval,inattrib,changerate) STORE_PURE_DELAY(delind,inval,outval,inattrib,changerate,'\0') 
#define DELAY_PROFILE_i(type,varoff,delind,initval,initgrowrate) INIT_PURE_DELAY(type,delind,varoff,1.0,initval,initgrowrate,'\0','\0') 
#define DELAY_PROFILE_a(delind,profile,inval,deltime) STORE_PURE_DELAY(delind,inval,0.0,deltime,0,profile)
#define DELAY_CONVEYOR_i(type,varoff,delind,initprofile,inittotal,initctime) INIT_PURE_DELAY(type,delind,varoff,initctime,inittotal,0,initprofile,'\0')
#define DELAY_CONVEYOR_a(delind,inval,deltime,drainage) STORE_PURE_DELAY(delind,inval,drainage,deltime,0,'\0')
#define DELAY_BATCH_i(type,varoff,delind,initbatch,inittime,initbacklog) INIT_PURE_DELAY(type,delind,varoff,inittime,initbatch,initbacklog,'\0','\0')
#define DELAY_BATCH_a(delind,inval,bsize,deltime) STORE_PURE_DELAY(delind,inval,bsize,deltime,0,'\0')
#define INTERNAL_RATE_OF_RETURN_i(type,varoff,delind,fperiod,initq,initt) INIT_PURE_DELAY(type,delind,varoff,fperiod,initq,initt,'\0','\0')
#define INTERNAL_RATE_OF_RETURN_a(delind,inval) STORE_PURE_DELAY(delind,inval,0,0,0,'\0')
#define NET_PRESENT_VALUE_i(type,varoff,delind,fperiod,fstart,foff) INIT_PURE_DELAY(type,delind,varoff,fperiod,fstart,foff,'\0','\0')

#define NET_PRESENT_VALUE_a(delind,inval,dr) STORE_PURE_DELAY(delind,inval,dr,0,0,'\0')
#define DEPRECIATE_BY_SCHEDULE_i(type,varoff,delind,sched,init,it,fstart) INIT_PURE_DELAY(type,delind,varoff,it,fstart,0,(TAB_TYPE *)sched,(TAB_TYPE *)init)
#define DEPRECIATE_BY_SCHEDULE_a(delind,inval) STORE_PURE_DELAY(delind,inval,0,0,0,'\0')
#define DEPRECIATE_STRAIGHTLINE_i(type,varoff,delind,deltime,fperiod,initq) INIT_PURE_DELAY(type,delind,varoff,deltime,initq,fperiod,'\0','\0')
#define DEPRECIATE_STRAIGHTLINE_a(delind,inval) STORE_PURE_DELAY(delind,inval,0,0,0,'\0')


CFUNCTION void* vensim_alloc_simmem(DATA_MATRIX *dm,CONSTANT_MATRIX *cm,V4BYTEU bytes) ;
CFUNCTION  BigREAL exp(BigREAL) ; 
CFUNCTION  BigREAL log(BigREAL) ; 
CFUNCTION  BigREAL sqrt(BigREAL) ; 
CFUNCTION  BigREAL sin(BigREAL) ; 
CFUNCTION  BigREAL sinh(BigREAL) ; 
CFUNCTION  BigREAL cos(BigREAL) ; 
CFUNCTION  BigREAL cosh(BigREAL) ; 
CFUNCTION  BigREAL tan(BigREAL) ; 
CFUNCTION  BigREAL tanh(BigREAL) ; 
CFUNCTION  BigREAL acos(BigREAL) ; 
CFUNCTION  BigREAL asin(BigREAL) ; 
CFUNCTION  BigREAL atan(BigREAL) ;


VECTOR_ARG * __cdecl vecarg_assign(GLOB_VARS *VGV,int ind,COMPREAL *vals)
{
   VGV->vector_arglist[ind].vals = vals ;
   return(VGV->vector_arglist + ind) ;
   }
#define vecarg_assign(ind,vals) vecarg_assign(VGV,ind,vals)

/********************************************
 inline functions - simply calls to vensim exports 
 ***************************************************/

#define ALLOCATE_AVAILABLE(allocation,request,request_p,available) (*VGV->ALLOCATE_AVAILABLE)(allocation,request,request_p,available)
#define MARKET_DS(ds,pprofile,price,isdemand) (*VGV->MARKET_DS)(ds,pprofile,price,isdemand)
#define FIND_MARKET_PRICE(demand_q,demand_p,supply_q,supply_p) (*VGV->FIND_MARKET_PRICE)(demand_q,demand_p,supply_q,supply_p)
#define MARKET_MATCH(sales,demandq,dshape,seller_attract,supplyq,sshape,buyer_priority,dsqconversion) (*VGV->MARKET_MATCH)(sales,demandq,dshape,seller_attract,supplyq,sshape,buyer_priority,dsqconversion)
#define VECTOR_SORT_ORDER(outvector,invector,d) (*VGV->VECTOR_SORT_ORDER)(outvector,invector,d)
#define VECTOR_RANK(outvector,invector,d) (*VGV->VECTOR_RANK)(outvector,invector,d)
#define VECTOR_REORDER(outvector,invector,sortvector) (*VGV->VECTOR_REORDER)(outvector,invector,sortvector)
#define GET_DATA_AT_TIME(dv,t) (*VGV->GET_DATA_AT_TIME)(dv,t)
#define GET_DATA_BETWEEN_TIMES(dv,t,mode) (*VGV->GET_DATA_BETWEEN_TIMES)(dv,t,mode)
#define GET_DATA_ATTRIBUTE(attrib,dv,t,mode) (*VGV->GET_DATA_ATTRIBUTE)(attrib,dv,t,mode)
#define GET_DATA_FIRST_TIME(dv) (*VGV->GET_DATA_FIRST_TIME)(dv)
#define GET_DATA_LAST_TIME(dv) (*VGV->GET_DATA_LAST_TIME)(dv)
#define GET_DATA_TOTAL_POINTS(dv) (*VGV->GET_DATA_TOTAL_POINTS)(dv)
#define GET_TIME_VALUE(relativeto,offset,measure) (*VGV->GET_TIME_VALUE)(relativeto,offset,measure)
#define MARKETP(request,priority,size, width,supply) (*VGV->MARKETP)(request,priority,size,width,supply)
#define MESSAGE(msg,severity) (*VGV->MESSAGE)(msg,severity)
#define LOOKUP_AREA(tab,start,end) (*VGV->LOOKUP_AREA)(tab,start,end)
#define LOOKUP_SLOPE(tab,x,mode) (*VGV->LOOKUP_SLOPE)(tab,x,mode)
#define LOOKUP_FORWARD(tab,in) (*VGV->LOOKUP_FORWARD)(tab,in)
#define LOOKUP_INVERT(tab,in) (*VGV->LOOKUP_INVERT)(tab,in)
#define LOOKUP_BACKWARD(tab,in) (*VGV->LOOKUP_BACKWARD)(tab,in)
#define LOOKUP_EXTRAPOLATE(tab,in) (*VGV->LOOKUP_EXTRAPOLATE)(tab,in)
#define TABLE(tab,in) (*VGV->TABLE)(tab,in)
#define RANDOM_NEGATIVE_BINOMIAL(xmin,xmax,p,n,shift,stretch,streamid) (*VGV->RANDOM_NEGATIVE_BINOMIAL)(xmin,xmax,p,n,shift,stretch,streamid)
#define RANDOM_NORMAL(xmin,xmax,mean,stddev,streamid) (*VGV->RANDOM_NORMAL)(xmin,xmax,mean,stddev,streamid)
#define RANDOM_WEIBULL(xmin,xmax,b,shift,stretch,streamid) (*VGV->RANDOM_WEIBULL)(xmin,xmax,b,shift,stretch,streamid)
#define RANDOM_BETA(xmin,xmax,a,b,shift,stretch,streamid) (*VGV->RANDOM_BETA)(xmin,xmax,a,b,shift,stretch,streamid)
#define RANDOM_BINOMIAL(xmin,xmax,p,n,shift,stretch,streamid) (*VGV->RANDOM_BINOMIAL)(xmin,xmax,p,n,shift,stretch,streamid)
#define RANDOM_TRIANGULAR(xmin,xmax,start,peak,stop,streamid) (*VGV->RANDOM_TRIANGULAR)(xmin,xmax,start,peak,stop,streamid)
#define RANDOM_EXPONENTIAL(xmin,xmax,shift,stretch,streamid) (*VGV->RANDOM_EXPONENTIAL)(xmin,xmax,shift,stretch,streamid)
#define RANDOM_LOOKUP(look,xmin,xmax,shift,stretch,streamid) (*VGV->RANDOM_LOOKUP)(look,xmin, xmax,shift,stretch,streamid)
#define RANDOM_POISSON(xmin,xmax,x, shift,stretch,streamid) (*VGV->RANDOM_POISSON)(xmin,xmax,x,shift,stretch,streamid)
#define RANDOM_GAMMA(xmin,xmax,x,shift,stretch,streamid) (*VGV->RANDOM_GAMMA)(xmin,xmax,x,shift,stretch,streamid)
#define RANDOM_UNIFORM(xmin,xmax,streamid) (*VGV->RANDOM_UNIFORM)(xmin,xmax,streamid)
#define RANDOM_0_1() (*VGV->RANDOM_0_1)()
#define QUEUE_AGE_AVERAGE(v,x) (*VGV->QUEUE_GETINFO)(v,x,0.0,F_QUEUE_AGE_AVERAGE)
#define QUEUE_AGE_OLDEST(v) (*VGV->QUEUE_GETINFO)(v,0.0,0.0,F_QUEUE_AGE_OLDEST)
#define QUEUE_AGE_IN_RANGE(v,low,hi) (*VGV->QUEUE_GETINFO)(v,low,hi,F_QUEUE_AGE_IN_RANGE)
#define QUEUE_ATTRIB_MAX(v) (*VGV->QUEUE_GETINFO)(v,0.0,0.0,F_QUEUE_ATTRIB_MAX)
#define QUEUE_ATTRIB_MIN(v) (*VGV->QUEUE_GETINFO)(v,0.0,0.0,F_QUEUE_ATTRIB_MIN)
#define QUEUE_ATTRIB_AVERAGE(v,x) (*VGV->QUEUE_GETINFO)(v,x,0.0,F_QUEUE_ATTRIB_AVERAGE)
#define QUEUE_ATTRIB_IN_RANGE(v,lo,hi) (*VGV->QUEUE_GETINFO)(v,lo,hi,F_QUEUE_ATTRIB_IN_RANGE)
#define QUEUE_ATTRIB_QUANTITY(v,x) (*VGV->QUEUE_GETINFO)(v,x,0.0,F_QUEUE_ATTRIB_QUANTITY)
#define INIT_PURE_DELAY(type,deloff,valoff,deltime,initval,missval,tab,tab2) (*VGV->INIT_PURE_DELAY)(type,deloff,valoff,deltime,initval,missval,tab,tab2)
#define INIT_SINTEG(type,deloff,valoff,initval,minval,maxval,quantval,specval,tolerance) (*VGV->INIT_SINTEG)(type,deloff,valoff,initval,minval,maxval,quantval,specval,tolerance)
#define STORE_PURE_DELAY(deloff,val,outval,deltime,changerate,tab) (*VGV->STORE_PURE_DELAY)(deloff,val,outval,deltime,changerate,tab)
#define COMPARE_SIMULT_VALUES(storeflag) (*VGV->COMPARE_SIMULT_VALUES)(storeflag)
#define FIND_ZERO(vx,vsx,vz,vsz,xini,fn,fns,tol,fmax_iter,method) (*VGV->FIND_ZERO)(vx,vsx,vz,vsz,xini,fn,fns,tol,fmax_iter,method)
#define INVERT_MATRIX(invmat,mat1,fn) (*VGV->INVERT_MATRIX)(invmat,mat1,fn)
#define VECTOR_LOOKUP(in,x,xmin,xmax,mode) (*VGV->VECTOR_LOOKUP)(in,x,xmin,xmax,mode)
#define external_function(val,nval,funcid) (*VGV->EXTERNAL_FUNCTION)(val,nval,funcid)
#define add_vecarg_info(stlist,n) (*VGV->add_vecarg_info)(stlist,n)
#define vselect_error(action,errtype,elmcount) (*VGV->vselect_error)(action,errtype,elmcount)


int VEFCC version_info(void)
 {
  return(COMPILED_VERSION) ;
  }

// this doesn't really "set" GV, it just checks
// in compiled sims, GV will be passed by reference in each function call
int VEFCC set_gv(GLOB_VARS *VGV) 
{
	
	if(!VGV )
		return 0;
	
	if ( VGV->vgv_magic_start != VGV_MAGIC_START )
		return 0;
	
	if ( VGV->vgv_magic_end != VGV_MAGIC_END)
		return 0 ;

	VGV->workspace = '\0'; // this will be allocated in the mdl_func that initializes time

	return 1 ;
}


/*************************************************************************
  functions supported in vensim
  all need to be BigREAL for correct argument passing to work
 *************************************************************************/
 
BigREAL SHIFT_IF_TRUE(GLOB_VARS *VGV,VECTOR_ARG *arr,BigREAL cond,BigREAL fnelm,BigREAL accum,
    BigREAL inival)
 {
  return (*VGV->SHIFT_IF_TRUE)(arr,cond,fnelm,accum,inival) ;
  }
#define SHIFT_IF_TRUE(arr,cond,fnelm,accum,inival) SHIFT_IF_TRUE(VGV,arr,cond,fnelm,accum,inival)

BigREAL ALLOC_P (BigREAL request,BigREAL priority,BigREAL width,BigREAL mp)
  {
   double half_width ;

   half_width = width/2 ;
  /* Consider degenerate cases of MP outside the priority distribution for
   the sector (which is a rectangle centered on PRIOR).  Otherwise just
   interpolate. */
  if ( mp >= priority + half_width)
      return(0.0) ;
  if ( mp <= priority - half_width )
     return(request) ;
  return(request * (priority + half_width - mp) / width) ;
  } /* alloc_p */


BigREAL ALLOCATE_BY_PRIORITY(GLOB_VARS *VGV,VECTOR_ARG *allocation,VECTOR_ARG *request,VECTOR_ARG *priority,int size,
      BigREAL width,BigREAL supply)
 {
  int i ;
  BigREAL mp ;
  mp = (*VGV->MARKETP)(request,priority,size,width,supply) ;
  for(i=0;i<size;i++)
     allocation->vals[i] = (COMPREAL)ALLOC_P(request->vals[i],priority->vals[i],width,mp) ;
  return(allocation->vals[0]) ;
  }
#define ALLOCATE_BY_PRIORITY(allocation,request,priority,size,width,supply) ALLOCATE_BY_PRIORITY(VGV,allocation,request,priority,size,width,supply)  

BigREAL QUANTUM(BigREAL a,BigREAL b)
  {
   BigREAL q ;
   long j ;
   if(b < 1.0E-6) return(a) ;
   q = a/b ;
   if(q > -2E9 && q < 2E9) {
      j = (long)q ;
      return(b*j) ;
      }
   return(a) ;
   }
   
BigREAL MODULO(BigREAL a,BigREAL b)
  {
   long j ;
   if(b < 1.0E-6) return(a) ;
   j = (long)(a/b) ;
   return(a - b*j) ;
   }
   
BigREAL STEP(GLOB_VARS *VGV, BigREAL height,BigREAL stime)
 {
  if(VGV->time_plus > stime)
     return(height) ;
  return(0.0) ;
  }
#define  STEP(height,stime) STEP(VGV,height,stime)


BigREAL RAMP(GLOB_VARS *VGV, BigREAL slope,BigREAL starttime,BigREAL stoptime)
 {
  if(VGV->time_plus > starttime) {
     if(VGV->time_plus < stoptime)
        return(slope*(VGV->time-starttime)) ;
     if(stoptime > starttime)
        return(slope*(stoptime - starttime)) ;
     return(slope*(VGV->time-starttime)) ;
     }
  return(0.0) ;
  }
#define RAMP(slope,starttime,stoptime) RAMP(VGV,slope,starttime,stoptime) 


BigREAL PULSE(GLOB_VARS *VGV, BigREAL ftime,BigREAL width) /* centered with time_plus */
  {
   if(width == 0)
      width = VGV->time_step ;
   if((VGV->time_plus > ftime) && (VGV->time_plus < (ftime + width)))
      return(1.0) ;
   return(0.0) ;
   }
#define PULSE(ftime,width) PULSE(VGV,ftime,width)   
 

BigREAL PULSE_TRAIN(GLOB_VARS *VGV, BigREAL ftime,BigREAL width,BigREAL repeat,BigREAL stop) /* centered with time_plus */
  {
   long i ;
   if(VGV->time_minus > stop || VGV->time_plus < ftime)
      return 0 ;
   if(repeat < VGV->time_step)
      return 1.0 ;
   i = (long)((VGV->time_plus - ftime)/repeat) ;
   ftime += i*repeat ;
   if(width == 0)
      width = VGV->time_step ;
   if(VGV->time_plus < (ftime + width))
      return(1.0) ;
   return(0.0) ;
   }
#define PULSE_TRAIN(ftime,width,repeat,stop) PULSE_TRAIN(VGV,ftime,width,repeat,stop)  
 
BigREAL POWER(BigREAL base,BigREAL expon)
 {
  int iexpon ; 
  BigREAL texpon ;
  
  iexpon = (int)expon ;
  texpon = iexpon ;
  if(texpon == expon) { /* integer exponent */ 
     texpon = 1.0 ;    
     if(iexpon < 0) {
        iexpon = -iexpon ;
        base = 1.0/base ;
        }  
     for(;iexpon>0;iexpon--)
        texpon *= base ; 
     return(texpon) ;
     }
  if(base == 0.0 && expon > 0.0)
      return(0.0) ;
  return(exp(expon*log(base))) ;
  }
  
BigREAL GAMMA_LN(BigREAL xx)
 {    
  double x,tmp,ser;
 static double cof[6]={76.18009173,-86.50532033,24.01409822,
    -1.231739516,0.120858003e-2,-0.536382e-5};
 int j;

 x=xx-1.0;
 tmp=x+5.5;
 tmp -= (x+0.5)*log(tmp);
 ser=1.0;
 for (j=0;j<6;j++) {
   x += 1.0;
   ser += cof[j]/x;
   }
  return -tmp+log(2.50662827465*ser);
  }
  

BigREAL TIME_BASE (GLOB_VARS *VGV, BigREAL intercept, BigREAL slope)
 {
  return(intercept + slope*VGV->time) ;
  }
#define  TIME_BASE(intercept,slope) TIME_BASE(VGV,intercept,slope)  

BigREAL SAMPLE_IF_TRUE(double oldval,double cond,double newval)
 {
  if(cond == 0.0)
     return(oldval) ;
  return(newval) ;
  }
  
BigREAL VECTOR_ELM_MAP(GLOB_VARS *VGV, VECTOR_ARG *in,BigREAL shift)
 {
  int ishift ; 
  COMPREAL *val ;
  charutf8 msg[512];

  if(shift < 0)
     ishift = (int)(shift - .5) ;  
  else
     ishift = (int)(shift + .5) ; 
  val = in->vals + ishift ;
  if(val < in->firstval ||
     val >= in->firstval + in->dim_info->u1.tot_vol) {
	  sprintf(msg, "VECTOR_ELM_MAP offset = %d outside array bounds (%d,%d) for input %s", ishift, (int)(in->firstval - in->vals), (int)(in->firstval + in->dim_info->u1.tot_vol - in->vals - 1), in->varname);
     (*VGV->error_message)(VERROR, msg);
#ifdef _DEBUG
	 sprintf(msg, "\tfirstval offset from LEVEL[0] %d", (int)(in->firstval - &((*VGV).LEVEL[0])));
     (*VGV->error_message)(VERROR, msg);
	 sprintf(msg, "\tin->vals offset from LEVEL[0] %d", (int)(in->vals - &((*VGV).LEVEL[0])));
     (*VGV->error_message)(VERROR, msg);
	 sprintf(msg, "\tin->dim_info->u1.tot_vol %d", (int)(in->dim_info->u1.tot_vol));
     (*VGV->error_message)(VERROR, msg);
	 sprintf(msg, "\tpointers: val %p, in->vals %p, in->firstval %p", val, in->vals, in->firstval);
     (*VGV->error_message)(VERROR, msg);
#endif
     return(NA) ;
     }
  return((BigREAL)(*val)) ;
  }
#define VECTOR_ELM_MAP(in,shift) VECTOR_ELM_MAP(VGV,in,shift)

int VEFCC check_struct_sizes(int *pInt)
{
#ifdef _DEBUG
	TABLE_RANGE_VDFX tmp;
	int nSize;
			
    if (sizeof(BLK_OFFSET) != 4)
    {
        *pInt = sizeof(BLK_OFFSET);
        return 4;
    }

    if (sizeof(VAL_OFF) != 8)
    {
        *pInt = sizeof(VAL_OFF);
        return 5;
    }
    
	#ifdef _VDFX
	nSize = sizeof(TABLE_RANGE_VDFX);
	if (nSize != 28)
	{
		*pInt = sizeof(tmp.history) ;
		*pInt = sizeof(tmp.first) ;
		*pInt = sizeof(tmp.tot_vol) ;
		*pInt = sizeof(tmp.next) ;

		*pInt = sizeof(TABLE_RANGE_VDFX);
		return 1;
	}
	#endif

	/*
     nSize = sizeof(TABLE_RANGE_VDF);
	if (nSize != 16)
	{
		*pInt = sizeof(TABLE_RANGE_VDF);
		return 2;
	}
     */

	#ifdef _VDFX
	nSize = sizeof(TAB_TYPE_VDFX);
	if (nSize != 64)
	{
		*pInt = sizeof(TAB_TYPE_VDFX);
		return 3;	
	}
	#endif



	/*
	nSize = sizeof(VECTOR_ARG);
	if (nSize != 32)
	{
		*pInt = sizeof(VECTOR_ARG);
		return 6;
	}
	*/
#endif //#ifdef _DEBUG

	return 0;
}

#pragma pack(pop) /* revert to normal - 8 for 64 bit operations */
